/*  WWW Utility Routines for NCSA httpd server
    Copyright 1994, Creative Internet Designs, All Rights Reserved
    Version 1.1

*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

   #define LF 10
   #define CR 13

   int include_file(char *flname);

   void get_host_info(char *hostinfo, int maxlength);
   void strip_end_cr(char *s);
   void strip_white(char *s);

   void unescape_url(char *url);
   void plus_to_space(char *s);

   char *makeword(char *line, char stop);
   char *fmakeword(FILE *f, char stop, int *len);
   char x2c(char *what);

/* ---------------------------------------------------------------------------
   include_file: includes the given file, returns non-zero if failure 
   ------------------------------------------------------------------------ */
   int include_file(char *flname) {

#define INC_BUF_LEN  2048

     int iret = 0;
     FILE *fp_inc;
     char buf[INC_BUF_LEN];

     if ((fp_inc = fopen(flname, "r")) == NULL) return 1;

     while (!(feof(fp_inc))) {

       fgets(buf, INC_BUF_LEN-1, fp_inc);
       if (ferror(fp_inc)) {iret = 1; break;}

       printf("%s", buf);

     }

     fclose(fp_inc);

     return iret;

   }

/* ---------------------------------------------------------------------------
   get_host_info: get info on remote user and host 
   ------------------------------------------------------------------------ */
   void get_host_info(char *hostinfo, int maxlength) {

    char *rh_user, *rh_name;

    strcpy(hostinfo, "");

    if ((rh_user = getenv("REMOTE_USER")) != NULL) {
      strncat(hostinfo, rh_user, maxlength-1);
      strcat(hostinfo, "@");
    } 

    if ((rh_name = getenv("REMOTE_HOST")) == NULL) {
      rh_name = getenv("REMOTE_ADDR");
    }    

    if (rh_name != NULL) {
      strncat(hostinfo, rh_name, maxlength-strlen(hostinfo));
    }
    
    /* if we didn't find any info, mark it "unknown" */
    if (strlen(hostinfo) == 0) strncpy(hostinfo, "Unknown", maxlength);
    
  }

/* ---------------------------------------------------------------------------
   strip_end_cr: strip lend carriage return /n, for "fgets" strings
   ------------------------------------------------------------------------ */
   void strip_end_cr(char *s) {

     for ( ; *s; s++) if (*s == '\n') *s = '\0';

   }  

/* ---------------------------------------------------------------------------
   strip_white: strip leading/trailing white space (TAB, CR, SPACE, etc.)
   ------------------------------------------------------------------------ */
   void strip_white(char *s) {

     int i = 0, j = 0;

     while (isspace(s[i])) i++;

     while (s[j++] = s[i++]);

     i = strlen(s) - 1;
     while (isspace(s[i])) s[i--] = '\0';

   }

/* ---------------------------------------------------------------------------
   Routines below based on "util.c" distributed with NCSA httpd. 
   ------------------------------------------------------------------------ */


/* ---------------------------------------------------------------------------
   makeword: split word out of line based on stop character
   ------------------------------------------------------------------------ */
   char *makeword(char *line, char stop) {
     
     int i = 0, j = 0;
     char *word = (char *) malloc(sizeof(char) * (strlen(line) + 1));

     /* loop until end or stop and dump in word */
     for (i=0; ((line[i]) && (line[i] != stop)); i++) word[i] = line[i];

     /* slap a null on the end of the word */
     word[i] = '\0';

     /* if not at end of line, increment to next char */
     if (line[i]) ++i;

     /* shift anything left in the string to the start */
     j = 0;
     while (line[j++] = line[i++]);

     return word;

   }

/* ---------------------------------------------------------------------------
   fmakeword:
   ------------------------------------------------------------------------ */
   char *fmakeword(FILE *f, char stop, int *cl) {

     int ll, wsize;
     char *word;

     wsize = 1024;
     ll = 0;

     word = (char *) malloc(sizeof(char) * (wsize + 1));

     while (1) {

       word[ll] = (char) fgetc(f);

       if (ll == wsize) {
         word[ll + 1] = '\0';
         wsize += 1024;
         word = (char *) realloc(word, sizeof(char) * (wsize+1));
       }

       --(*cl);

       if ((word[ll] == stop) || (feof(f)) || (!(*cl))) {

         if(word[ll] != stop) ll++;
           word[ll] = '\0';
           return word;
        }

        ++ll;

      }

   }

/* ---------------------------------------------------------------------------
   x2c: convert double digit hex code to a decimal character value 
   ------------------------------------------------------------------------ */
   char x2c(char *what) {

     register char digit;

     digit  = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
     digit *= 16;
     digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
     
     return(digit);

   }

/* ---------------------------------------------------------------------------
   unescape_url: convert back escaped characters (like %22 for example)
   ------------------------------------------------------------------------ */
   void unescape_url(char *url) {

     register int i, j;

     for (i=0,j=0; url[j]; ++i,++j) {

       url[i] = url[j];

       if (url[i] == '%') {

         /* convert the decimal string to a character */
         url[i] = x2c(&url[j+1]);

         /* advance y past the escaped char code */
         j += 2;

       }
    }

    /* mark end fo the string we are building */
    url[i] = '\0';

   }

/* ---------------------------------------------------------------------------
   plus_to_space: convert plus signs to spaces in form results
   ------------------------------------------------------------------------ */
   void plus_to_space(char *s) {

     for ( ; *s; s++) if (*s == '+') *s = ' ';

  }

/* end */

